#!/usr/bin/env python
# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Revision $Id: test_genmsg_py.py 4140 2009-04-10 23:14:00Z sfkwc $

import roslib; roslib.load_manifest('test_rospy')

import os
import sys
import struct
import unittest
from cStringIO import StringIO
import time
import random
import math

from roslib.message import SerializationError
        
class TestGenmsgPy(unittest.TestCase):

    ## Utility for testing roundtrip serialization
    ## @param orig Message to test roundtrip serialization of
    ## @param blank Uninitialized instance of message to deserialize into
    ## @param float bool: if True, use almostEquals instead of equals
    ## comparison. This variant assumes only data field is named
    ## 'data'
    def _test_ser_deser(self, orig, blank, float=False):
        b = StringIO()
        orig.serialize(b)
        blank.deserialize(b.getvalue())
        if not float:
            self.assertEquals(orig, blank, str(orig)+" != "+str(blank))
        else:
            self.assertAlmostEquals(orig.data, blank.data, 5)

    def test_std_msgs_empty(self):
        from std_msgs.msg import Empty
        self.assertEquals(Empty(), Empty())
        self._test_ser_deser(Empty(), Empty())

    def test_std_msgs_String(self):
        from std_msgs.msg import String
        self.assertEquals(String(), String())
        self.assertEquals('', String().data)
        # default value should be empty string
        self.assertEquals(String(''), String())
        self.assertEquals(String(''), String(''))
        self.assertEquals(String('foo'), String('foo'))
        self.assertEquals(String('foo'), String(data='foo'))
        self.assertEquals(String(data='foo'), String(data='foo'))
        
        self.assertNotEquals(String('foo'), String('bar'))
        self.assertNotEquals(String('foo'), String(data='bar'))
        self.assertNotEquals(String(data='foo'), String(data='bar'))
        
        self._test_ser_deser(String(''), String())
        self._test_ser_deser(String('a man a plan a canal panama'), String())

    def test_std_msgs_SignedInt(self):
        from std_msgs.msg import Byte, Int8, Int16, Int32, Int64
        for cls in [Byte, Int8, Int16, Int32, Int64]:
            v = random.randint(1, 127)
            self.assertEquals(cls(), cls())
            self.assertEquals(0, cls().data)
            self.assertEquals(cls(), cls(0))
            self.assertEquals(cls(0), cls(0))        
            self.assertEquals(cls(v), cls(v))
            self.assertEquals(cls(-v), cls(-v))
            self.assertEquals(cls(v), cls(data=v))        
            self.assertEquals(cls(data=v), cls(data=v))
        
            self.assertNotEquals(cls(v), cls())
            self.assertNotEquals(cls(data=v), cls(data=-v))
            self.assertNotEquals(cls(data=v), cls(data=v-1))            
            self.assertNotEquals(cls(data=v), cls(v-1))
            self.assertNotEquals(cls(v), cls(v-1))
            
            self._test_ser_deser(cls(), cls())
            self._test_ser_deser(cls(0), cls())
            self._test_ser_deser(cls(-v), cls())
            self._test_ser_deser(cls(v), cls())

        # rospy currently does not spot negative overflow due to the fact that Python's struct doesn't either
        widths = [(8, Byte), (8, Int8), (16, Int16), (32, Int32), (64, Int64)]
        for w, cls in widths:
            maxp = long(math.pow(2, w-1)) - 1
            maxn = -long(math.pow(2, w-1)) + 1
            self._test_ser_deser(cls(maxp), cls())
            self._test_ser_deser(cls(maxn), cls())
            try:
                cls(maxp+1)._check_types()
                self.fail("check_types should have noted width error[%s]: %s, %s"%(w, maxp+1, cls.__name__))
            except SerializationError: pass
            try:
                cls(maxn-1)._check_types()
                self.fail("check_types should have noted width error[%s]: %s, %s"%(w, maxn-1, cls.__name__))
            except SerializationError: pass
            
    def test_std_msgs_UnsignedInt(self):
        from std_msgs.msg import Char, UInt8, UInt16, UInt32, UInt64
        for cls in [Char, UInt8, UInt16, UInt32, UInt64]:
            v = random.randint(1, 127)
            self.assertEquals(cls(), cls())
            self.assertEquals(0, cls().data)
            self.assertEquals(cls(), cls(0))
            self.assertEquals(cls(0), cls(0))        
            self.assertEquals(cls(v), cls(v))
            self.assertEquals(cls(v), cls(data=v))        
            self.assertEquals(cls(data=v), cls(data=v))
        
            self.assertNotEquals(cls(v), cls())
            self.assertNotEquals(cls(data=v), cls(data=-v))
            self.assertNotEquals(cls(data=v), cls(data=v-1))            
            self.assertNotEquals(cls(data=v), cls(v-1))
            self.assertNotEquals(cls(v), cls(v-1))
            
            self._test_ser_deser(cls(), cls())
            self._test_ser_deser(cls(0), cls())
            self._test_ser_deser(cls(v), cls())

            try:
                cls(-1)._check_types()
                self.fail("check_types should have noted sign error[%s]: %s"%(w, cls.__name__))
            except SerializationError: pass

        # rospy currently does not spot negative overflow due to the fact that Python's struct doesn't either
        widths = [(8, Char), (8, UInt8), (16, UInt16), (32, UInt32), (64, UInt64)]
        for w, cls in widths:
            maxp = long(math.pow(2, w)) - 1
            self._test_ser_deser(cls(maxp), cls())
            try:
                cls(maxp+1)._check_types()
                self.fail("check_types should have noted width error[%s]: %s, %s"%(w, maxp+1, cls.__name__))
            except SerializationError: pass
            
    def test_std_msgs_Float(self):
        from std_msgs.msg import Float32, Float64
        for cls in [Float32, Float64]:
            self.assertEquals(cls(), cls())
            self.assertEquals(0., cls().data)
            self.assertEquals(cls(), cls(0.))
            self.assertEquals(cls(0.), cls(0.))        
            self.assertEquals(cls(1.), cls(1.))
            self.assertEquals(cls(1.), cls(data=1.))        
            self.assertEquals(cls(data=1.), cls(data=1.))
            self.assertEquals(cls(math.pi), cls(math.pi))
            self.assertEquals(cls(math.pi), cls(data=math.pi))        
            self.assertEquals(cls(data=math.pi), cls(data=math.pi))
        
            self.assertNotEquals(cls(1.), cls())
            self.assertNotEquals(cls(math.pi), cls())
            self.assertNotEquals(cls(data=math.pi), cls(data=-math.pi))
            self.assertNotEquals(cls(data=math.pi), cls(data=math.pi-1))            
            self.assertNotEquals(cls(data=math.pi), cls(math.pi-1))
            self.assertNotEquals(cls(math.pi), cls(math.pi-1))
            
            self._test_ser_deser(cls(), cls())
            self._test_ser_deser(cls(0.), cls())
            self._test_ser_deser(cls(1.), cls(), float=True)
            self._test_ser_deser(cls(math.pi), cls(), float=True)

    def test_std_msgs_MultiArray(self):
        # multiarray is good test of embed plus array type
        from std_msgs.msg import Int32MultiArray, MultiArrayDimension, MultiArrayLayout
        
        dims = [MultiArrayDimension('foo', 1, 2), MultiArrayDimension('bar', 3, 4),\
                    MultiArrayDimension('foo2', 5, 6), MultiArrayDimension('bar2', 7, 8)]
        for d in dims:
            self.assertEquals(d, d)

        self.assertEquals(Int32MultiArray(),Int32MultiArray())
        self.assertEquals(Int32MultiArray(layout=MultiArrayLayout()),Int32MultiArray())
        self.assertEquals(Int32MultiArray(layout=MultiArrayLayout(), data=[1, 2, 3]),Int32MultiArray(data=[1, 2, 3]))
        self.assertEquals(Int32MultiArray(layout=MultiArrayLayout(), data=[1, 2, 3]),\
                              Int32MultiArray(layout=MultiArrayLayout(),data=[1, 2, 3]))
        self.assertEquals(Int32MultiArray(layout=MultiArrayLayout(dim=[]), data=[1, 2, 3]),\
                              Int32MultiArray(layout=MultiArrayLayout(),data=[1, 2, 3]))
        self.assertEquals(Int32MultiArray(layout=MultiArrayLayout([], 0), data=[1, 2, 3]),\
                              Int32MultiArray(layout=MultiArrayLayout(),data=[1, 2, 3]))
        self.assertEquals(Int32MultiArray(layout=MultiArrayLayout(dim=[], data_offset=0), data=[1, 2, 3]),\
                              Int32MultiArray(layout=MultiArrayLayout(),data=[1, 2, 3]))
        self.assertEquals(Int32MultiArray(layout=MultiArrayLayout(dim=dims, data_offset=0), data=[1, 2, 3]),\
                              Int32MultiArray(layout=MultiArrayLayout(dim=dims),data=[1, 2, 3]))
        self.assertEquals(Int32MultiArray(layout=MultiArrayLayout(dims, 10), data=[1, 2, 3]),\
                              Int32MultiArray(layout=MultiArrayLayout(dim=dims,data_offset=10),data=[1, 2, 3]))


        self.assertNotEquals(Int32MultiArray(data=[1, 2, 3]),Int32MultiArray(data=[4,5,6]))
        self.assertNotEquals(Int32MultiArray(layout=MultiArrayLayout([], 1), data=[1, 2, 3]),\
                                 Int32MultiArray(layout=MultiArrayLayout([], 0),data=[1, 2, 3]))
        self.assertNotEquals(Int32MultiArray(layout=MultiArrayLayout([], 1), data=[1, 2, 3]),\
                                 Int32MultiArray(layout=MultiArrayLayout(dim=[]),data=[1, 2, 3]))
        self.assertNotEquals(Int32MultiArray(layout=MultiArrayLayout(dims, 10), data=[1, 2, 3]),\
                                 Int32MultiArray(layout=MultiArrayLayout(dim=dims,data_offset=11),data=[1, 2, 3]))
        self.assertNotEquals(Int32MultiArray(layout=MultiArrayLayout(dim=dims, data_offset=10), data=[1, 2, 3]),\
                                 Int32MultiArray(layout=MultiArrayLayout(dim=dims[1:],data_offset=10),data=[1, 2, 3]))
        

        self._test_ser_deser(Int32MultiArray(),Int32MultiArray())
        self._test_ser_deser(Int32MultiArray(layout=MultiArrayLayout()),Int32MultiArray())
        self._test_ser_deser(Int32MultiArray(data=[1, 2, 3]),Int32MultiArray())

    
if __name__ == '__main__':
    import rostest
    rostest.unitrun('test_rospy', sys.argv[0], TestGenmsgPy)
